function Dup = disjoin_cones(D, Dup, rad, min_dist)
% Ensure that cones of new dictionary are disjoint, by moving them towards
% the old dictionary when necessary.
% If some cones of the old dictionary are too close, the corresponding new
% cones keep their values or take the old ones, if the distance between them
% is larger or smaller, respectively, than the old distance.
% Input:
%   D        - old dictionary (central atoms); it is assumed that the cones
%            generated by D are disjoint
%   Dup      - updated dictionary (central atoms)
%   rad      - cone radii (if scalar, then all atoms have the same radius)
%   min_dist - minimum distance between cones ("zero" = the sum of radii)
%
% Output:
%   Dup    - updated dictionary with disjoint atoms

% BD 31.07.2022

if nargin < 4
  min_dist = 0;
end

tol = 1e-12;

[~,n] = size(D);

if length (rad) == 1
  rad = rad * ones(n,1);
end

% align atoms (in case they are opposed)
for i = 1:n
  if D(:,i)'*Dup(:,i) < 0
    Dup(:,i) = -Dup(:,i);
  end
end

% check if there are cones too close
s_list = check_cone_superposition(Dup, rad, min_dist);
it_nr = 0;
old_too_close_list = [];    % list of pairs in the old dictionary that are too close
while ~isempty(s_list)
    for k = 1 : size(s_list,1)
      i1 = s_list(k,1);
      i2 = s_list(k,2);    % indices of atoms that are too close
      if D(:,i1)'*D(:,i2) < 0
        D(:,i2) = - D(:,i2);
      end
      if Dup(:,i1)'*Dup(:,i2) < 0
        D(:,i2) = - D(:,i2);
      end

      %d = norm(D(:,i1) - D(:,i2)); % distance between old atoms
      %dup = norm(Dup(:,i1) - Dup(:,i2)); % distance between new atoms
      d = sqrt(2 - 2*abs(D(:,i1)'*D(:,i2))); % distance between old atoms
      dup = sqrt(2 - 2*abs(Dup(:,i1)'*Dup(:,i2))); % distance between new atoms
      dok = rad(i1)*sqrt(1 - rad(i2)*rad(i2)/4) + rad(i2)*sqrt(1 - rad(i1)*rad(i1)/4) + min_dist; % desired distance 
      if d + tol < dok    % check whether D respected the distance rule
        if it_nr == 0   % action is needed only at the first iteration
          if dup < d    % the distance has decreased, keep the old dictionary
            Dup(:,i1) = D(:,i1);
            Dup(:,i2) = D(:,i2);
          end           % otherwise, take the new atoms as they are
          % update the list of atoms that are too close and cannot be moved
          old_too_close_list = [old_too_close_list; s_list(k,:)];
        end
      else           % move towards old atoms as much as necessary
        if 0 % successive approximation
            for ii = 1 : 5 % approximate, using proportionality on straight lines (not on sphere)
              f = (d - dok) / (d - dup);  % fraction of distance with respect to old atom
              Dup(:,i1) = normc((1-f)*D(:,i1) + f*Dup(:,i1));
              Dup(:,i2) = normc((1-f)*D(:,i2) + f*Dup(:,i2));
              dup = norm(Dup(:,i1) - Dup(:,i2));
            end
        else
            % bisection algorithm
            fi = 0;
            ff = 1;
            d1i = D(:,i1);
            d1f = Dup(:,i1);
            d2i = D(:,i2);
            d2f = Dup(:,i2);
            for ii = 1:12
              f = (fi+ff)/2;
              d1 = normc((1-f)*d1i + f*d1f);
              d2 = normc((1-f)*d2i + f*d2f);
              % if norm(d1-d2) < dok
              if sqrt(2-2*abs(d1'*d2)) < dok
                ff=f;
                d1f = d1;
                d2f = d2;
              else
                fi = f;
                d1i = d1;
                d2i = d2;
              end
            end
            Dup(:,i1) = d1i;
            Dup(:,i2) = d2i;
        end
      end
    end
    s_list = check_cone_superposition(Dup, rad, min_dist);   % check again after solving, maybe new conflicts appeared

    if it_nr > 10
      s_list
    end
    
    if isempty(s_list)  % cut it short if dictionary is ok
      break
    end
    
    % check whether the list contains only old atoms that are too close
    if ~isempty(old_too_close_list)
      only_old = 1;
      for k = 1 : size(s_list,1)
        if ~ismember(s_list(k,:), old_too_close_list, 'rows')  % not old
          only_old = 0;
          break
        end
      end
      if only_old  % no more solvable conflicts
        break
      end
    end
    
    it_nr = it_nr+1;
end
