function [D, X, err] = dl_cone_paksvd_opt(Y, D, s, iternum, rad, iter_cd, err_thresh, ...
                                          min_d_cones, update_quota, q_dec_rate)
% Cone dictionary learning. Dictionary update is made in Parallel AK-SVD style,
% using an optimal update of the atoms.
% The radii of the cones are not changed.
%
% Input:
%   Y           - set of signals
%   D           - dictionary (normalized atoms)
%   s           - sparsity level
%   iternum     - number of learning iterations
%   rad         - cone radii (if scalar, then all atoms have the same radius)
%   iter_cd     - number of iterations in the coordinate descent refinement
%                 (default = 1)
%   err_thresh  - representation error for which OMP is stopped
%   min_d_cones - minimum distance between cones
%   update_quota- fraction of atoms that are updated in one pass over data (default = 1)
%   q_dec_rate  - quota decrease rate (update_quota *= q_dec_rate in each iteration)
%
% Output:
%   D           - trained dictionary
%   X           - representation matrix
%   err         - vector of errors at each iteration

% BD 24.07.2022

if nargin < 8
  min_d_cones = 0;
end
if nargin < 9
  update_quota = 1;
end
if nargin < 10
  q_dec_rate = 1;
end

[m,n] = size(D);
N = size(Y,2);

err = zeros(1,iternum);
for k = 1 : iternum
  n_update = round(update_quota*n);
  X = zeros(n,N);
  Dav = zeros(m,n);
  e = 0;
  if n_update < n
    crt_atoms = randperm(n);
    crt_atoms = crt_atoms(1:n_update);
  end
  for i = 1 : N
    % compute representation of a signal
    y = Y(:,i);
    [x, Da, support] = omp_cone_single(y, D, s, rad, iter_cd, err_thresh);
    X(support,i) = x;
    r = y - Da*x;       % representation residual
    if n_update == n  % fully parallel case (don't waste time with column computation)
      for j = 1 : length(support)
        f = r + Da(:,j)*x(j);
        Dav(:,support(j)) = Dav(:,support(j)) + x(j)*(f - x(j)*(Da(:,j) - D(:,support(j))));
      end
    else
      for j = 1 : length(support)
        if any(support(j)==crt_atoms)
          f = r + Da(:,j)*x(j);
          Dav(:,support(j)) = Dav(:,support(j)) + x(j)*(f - x(j)*(Da(:,j) - D(:,support(j))));
        end
      end
    end
    e = e + norm(r)^2;
  end
  e = sqrt(e)/sqrt(m*N);
  err(k) = e;

  % if new atoms are zero, replace with old atoms
  for j = 1:n
    if norm(Dav(:,j)) == 0
      Dav(:,j) = D(:,j);
    end
  end
  
  % new atoms
  D = disjoin_cones(D, normc(Dav), rad, min_d_cones);
  %min_dist_dict(D, rad)
  
  update_quota = update_quota * q_dec_rate;
end
