% comparison between Cone-DL and AK-SVD+Cone-OMP
% for data from ADBench (but not only)

% BD 26.06.2023

datasets_sort_m = {'musk', 'SpamBase', 'landsat', 'satellite', 'satimage', ...
                   'ionosphere', 'WPBC', 'letter', 'WDBC', 'fault', ...
                   'cardio', 'Waveform', 'Hepatitis', 'Lymphography', 'pendigits', ...
                   'wine', 'vowels', 'cover', 'magic.gamma.mat', 'breastw', ...
                   'Stamps', 'WBC', 'yeast', 'glass', 'annthyroid', ...
                   'mammography', 'Pima', 'thyroid', 'vertebral', 'Wilt'};

nf = 1;         % number of dataset in the list above
data_file_name = ['../DATA_DEPENDENCY/dependency_', datasets_sort_m{nf}];

n_tests = 10; % number of tests on which the average is made

% parameters and sizes
overcompleteness = 3;
s = 3;
rad_max = 0.2;
rad_min = 0.01;
iternum = 200;
iter_cd = 10;
err_thresh = 1e-7;
min_dist_cones = 0.01;
init_dl = 0;
iter_init = 10;     % iterations for DL initialization

alg_legend = {'Cone-DL', 'Cone-DL + swap', 'AK-SVD + Cone-OMP', 'AK-SVD + Cone-OMP + swap'};
    
% check if data file already exists
fig_data_name = ['./figs/fig_data_', datasets_sort_m{nf}, '_s_', num2str(s), ...
           '_c_', num2str(overcompleteness), '_r_', num2str(rad_min), '_', num2str(rad_max), '.mat'];
if isfile(fig_data_name)
    load(fig_data_name)
else % compute if data do not exist

    % data loading and normalization
    fprintf("%s\n", data_file_name)
    load(data_file_name)
    Y=Y';
    [m,N] = size(Y);
    vm = mean(Y,2);             % Z-score normalization
    Y = double(Y) - repmat(vm,1,N);         % subtract mean
    sm = std(Y,0,2);
    Y = Y ./ repmat(sm,1,N);
    
    n = overcompleteness * m;
    
    err1_av = 0;
    err2_av = 0;
    err3_av = 0;
    err4_av = 0;
    for it = 1 : n_tests
      dist_ok = 0;
      i0 = 0;
      while ~dist_ok
        % generate radii
        rad = rad_min + (rad_max-rad_min) * rand(n,1);
        D0 = randn(m,n);
        D0 = normc(D0);  % initial dictionary
        if init_dl
          D0 = DL(Y, D0, s, iter_init);
        end
        if isempty(check_cone_superposition(D0, rad))
          dist_ok = 1;
        end
        i0 = i0+1;
        if i0 == 10
          break
        end
      end
      if i0==10
        error('D0 cones superposition')
      end
    
      % dictionary learning
      % Cone-DL
      [D1, ~, err1] = dl_cone_paksvd_opt(Y, D0, s, iternum, rad, iter_cd, err_thresh, min_dist_cones);
      err1_av = err1_av + err1;
      % swap radii by use
      [~, Xs] = compute_rep_cone(D1, Y, s, rad, iter_cd, err_thresh);
      atom_use = sum(squeeze(Xs)~=0,2);
      [~,Ip] = sort(atom_use,'descend');
      [~,Ir] = sort(rad,'descend');
      rad2 = zeros(n,1);
      rad2(Ip) = rad(Ir);  % swap original radii!!!
      rad2 = adjust_radii(D1, rad2, min_dist_cones);
    
      err2 = 0;
      for i = 1 : N
        y = Y(:,i);
        [x, Da, support] = omp_cone_single(y, D1, s, rad2, iter_cd, err_thresh);
        err2 = err2 + norm(y - Da*x)^2;
      end
      err2 = sqrt(err2)/sqrt(m*N);
      err2_av = err2_av + err2;
    
      % AK-SVD
      update = 'aksvd';
      replatom = 'random';
      D3 = DL(Y, D0, s, iternum, str2func(update), {}, 'replatom', replatom); % standard DL 
      % run Cone-OMP with initial radii
      rad3 = adjust_radii(D3, rad, min_dist_cones);  % ensure that cones are disjoint for AK-SVD
    
      err3 = 0;
      for i = 1 : N
        y = Y(:,i);
        [x, Da, support] = omp_cone_single(y, D3, s, rad3, iter_cd, err_thresh);
        err3 = err3 + norm(y - Da*x)^2;
      end
      err3 = sqrt(err3)/sqrt(m*N);
      err3_av = err3_av + err3;
      
      % swap radii by use
      [~, Xs] = compute_rep_cone(D3, Y, s, rad3, iter_cd, err_thresh);
      atom_use = sum(squeeze(Xs)~=0,2);
      [~,Ip] = sort(atom_use,'descend');
      [~,Ir] = sort(rad3,'descend');
      rad_swap = zeros(n,1);
      rad_swap(Ip) = rad(Ir);  % swap original radii!!!
      rad_swap = adjust_radii(D3, rad_swap, min_dist_cones);
    
      err4 = 0;
      for i = 1 : N
        y = Y(:,i);
        [x, Da, support] = omp_cone_single(y, D3, s, rad_swap, iter_cd, err_thresh);
        err4 = err4 + norm(y - Da*x)^2;
      end
      err4 = sqrt(err4)/sqrt(m*N);
      err4_av = err4_av + err4;
    end
    
    best_err = [min(err1_av) err2_av err3_av err4_av] / n_tests;
    % show sorted best errors
    for i = 1 : length(alg_legend)
      fprintf("%24s: %8.6f\n", alg_legend{i}, best_err(i));
    end
    
    % save data
    save(fig_data_name, "err1_av", "err2_av", "err3_av", "err4_av", "n_tests")
end

% plot
font_size = 12;
line_width = 2;
figure, hold on
plot(log10(err1_av/n_tests), 'LineWidth', line_width)
plot(log10(err2_av/n_tests*ones(1,iternum)), 'LineWidth', line_width)
plot(log10(err3_av/n_tests*ones(1,iternum)), 'LineWidth', line_width)
plot(log10(err4_av/n_tests*ones(1,iternum)), 'LineWidth', line_width)
legend(alg_legend, 'FontSize', font_size)
xlabel('iteration', 'FontSize', font_size)
ylabel('log10(error)', 'FontSize', font_size)
title(datasets_sort_m{nf}, 'FontSize', font_size)
grid
hold off

fig_name = ['./figs/fig_', datasets_sort_m{nf}, '_s_', num2str(s), '_c_', num2str(overcompleteness), ...
           '_r_', num2str(rad_min), '_', num2str(rad_max), '.eps'];
print(fig_name, '-depsc')

